#!/usr/bin/perl

##
# tomcat_stats.pl - Get JMX data via tomcat's jmxproxy.
# Written by Pontus Fuchs 2009. pontus.fuchs@tactel.se
#
# This script can be used as a Data Query or as a Data Input method. See examples in usage.
#
# Version 0.8
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##


require HTTP::Request;
require LWP::UserAgent;
use MIME::Base64;
use strict;


# Edit these to fit your systems

my $user = "admin";
my $pass = "admin";
my $proto = "http://";
my $jmxurl = "/manager/jmxproxy/?qry=";
my $port = 80;
my $realm = "Tomcat Manager Application";
my $cacheDir = "/tmp";
my $cacheAge = 30;	#In seconds

#my $cacheDir;	#Set to empty to disable cache (not recommended)
#my $debugLog = "/tmp/tomcat_cacti.log";
my $debugLog;	#Set to emtpy to disable debug log

# Don't edit below

	my $host = $ARGV[0];

	my $debugArgs = "";
	foreach my $arg (@ARGV)
	{
		$debugArgs .= $arg . " ";
	}
	debug("Args: $debugArgs\n");	

	if(@ARGV < 3)
	{
		usage();
		exit(1);
	}

	my $mode = $ARGV[1];

	my $qry;

	if($mode eq "tp")
	{
		$qry = "Catalina:type=ThreadPool,name=*";
	}
	elsif($mode eq "gc")
	{
		$qry = "java.lang:type=GarbageCollector,name=*";
	}
	elsif($mode eq "mp")
	{
		$qry = "java.lang:type=MemoryPool,name=*";
	}
	elsif($mode eq "wm")
	{
		$qry = "Catalina:j2eeType=WebModule,name=*,J2EEApplication=none,J2EEServer=none";
	}
	elsif($mode eq "sv")
	{
		$qry = "Catalina:j2eeType=Servlet,name=*,WebModule=*,J2EEApplication=none,J2EEServer=none";
	}
	elsif($mode eq "datainput")
	{
		datainput();
		exit;
	}
	else
	{
		$qry = $mode;
	}


	if($ARGV[2] eq "index")
	{
		my $objs = doQuery($qry);
		foreach my $k (keys(%$objs))
		{
			print $k."\n";
		}
	}
	elsif($ARGV[2] eq "debug")
	{
		my $objs = doQuery($qry);

		foreach my $k (keys (%$objs))
		{
			print "$k\n";
			print "=================\n";
			my $o = $objs->{$k};
			foreach my $key (keys(%$o))
			{
				print $key . ":" . $o->{$key} . "\n";
			}
			print "\n";
		}

	}
	elsif($ARGV[2] eq "query")
	{
		my $attr = $ARGV[3];
		my $objs = doQuery($qry);

		foreach my $k (keys (%$objs))
		{
			my $o = $objs->{$k};
			my $a;
			if(not defined $o->{$attr})
			{
				$a = "Not found!";
			}
			else
			{
				$a = $o->{$attr};
			}

			print $k . "!" . $a . "\n";
		}

	}
	elsif($ARGV[2] eq "get")
	{
		my $attr = $ARGV[3];
		my $index = $ARGV[4];

		if($attr eq "" or $index eq "")
		{
			die("index och attribute missing");
		}

		my $objs = doQuery($qry);
		if(not defined $objs->{$index})
		{
			print "$index not found\n";
		}
		if(not defined $objs->{$index}->{$attr})
		{
			print "$attr not found\n";
		}
		print $objs->{$index}->{$attr} ."\n";

	}
	else
	{
		die("Uknown cmd");
	}


sub debug
{
	my $s = shift;

	#Debug
	if(defined($debugLog))
	{
		open(LOG, ">>$debugLog");
		print LOG localtime() . ": " . $s;
		close(LOG);
	}
}


#
# Print usage
#
sub usage
{
	print "Usage: $0 hostname mode cmd (data)\n";
	print "  hostname - Host to query\n";
	print "  mode     - gc, tp wm, sv or qustom JMX path\n";
	print "           - gc Garbage collectors\n";
	print "           - tp Thread pools\n";
	print "           - mp Mempry Pools\n";
	print "           - wm Web Modules\n";
	print "           - sv Servlets\n";
	print "\n";
	print "  cmd      - index, query, get or debug\n";
	print "\n";
	print "Example get GCs       : $0 localhost gc index\n";
	print "Example query GCs     : $0 localhost gc query Name\n";
	print "Example get GC data   : $0 localhost gc get CollectionCount java.lang:type=GarbageCollector,name=MarkSweepCompact (Note that you GC might have a different name)\n";
	print "\n";
	print "Example index JMX URL : $0 localhost Catalina:type=Connector,* index\n";
	print "Example query JMX URL : $0 localhost Catalina:type=Connector,* query protocol\n";
	print "Example get JMX URL   : $0 localhost Catalina:type=Connector,* get port Catalina:type=Connector,port=8080\n";
	print "\n";
	print "Example Data input    : $0 localhost datainput java.lang:type=Threading ThreadCount PeakThreadCount\n";
}

#
# Use as a data input script
#
sub datainput
{
	if(@ARGV < 4)
	{
		usage();
		exit 1;
	}

	my $qry = $ARGV[2];
	my $objects = doQuery($qry);

	if(%$objects != 1)
	{
		die("query returned multiple objects\n");
	}
	my $key = (keys(%$objects))[0];
	debug("object key: $key");;
	my $object = $objects->{$key};
	
	for(my $i = 3; $i < @ARGV; $i++)
	{
		my $attr = $ARGV[$i];
		print $attr. ":" . $object->{$attr};
		if(@ARGV - 1 > $i)
		{
			print " ";
		}
	}
}


#
# Perform an jmx query and return certain 
# parameter qry: jmx query
# attrs: Array of attributes to look for
#
sub getAttributes
{
	my $qry = shift;
	my $attrs = shift;

	my $data = doQuery($qry);
	my @lines = split("\n", $data);

	my @out;
	foreach my $attr (@$attrs)
	{
		foreach my $line (@lines)
		{
			if($line =~ /^$attr: (.+)/)
			{
				push(@out, "$attr:$1");
			}
		}
	}
	return \@out;
}

#
# Perform an jmx query.
# parameter qry: jmx query
# Return: Hash of jmx objects. The object name is used as keys.
# The jmx object is a hash with the attributes as keys.
#
sub doQuery
{
	my $qry = shift;
	my $cacheName = $cacheDir . "/" . $host ."-cacti_tomcat-".  encode_base64($qry);
	chomp($cacheName);
	my $body;

	#Try to fetch from cache
	if(defined($cacheDir))
	{
		if(-f $cacheName)
		{
			my @stats = stat($cacheName);
			my $modTime = $stats[9];
			if($modTime + $cacheAge > time())
			{
				debug("Reading from cache: $cacheName\n");
				open(CACHE, "<$cacheName");
				$body = do { local $/; <CACHE> };
				close(CACHE);
			}
		}
		
	}

	if($body eq "")
	{
		#Get via HTTP

		my $url = $proto.$host.$jmxurl.$qry;

		debug("Doing HTTP request:". $url . "\n");
		my $request = HTTP::Request->new(GET => $url);

		my $ua = LWP::UserAgent->new;
		$ua->credentials("$host:$port", $realm, $user, $pass);

		my $response = $ua->request($request);
		if($response->code() != 200)
		{
			die("HTTP Reponse: " .$response->status_line);
		}
		$body = $response->content;
		$body =~ s/\r//g;

		#Put info cache
		if(defined($cacheDir))
		{
			debug("Writing response to cache: $cacheName\n");
			open(CACHE, ">$cacheName") or die ("Could not open cache $cacheName");
			print CACHE $body;
			close(CACHE);
		}

	}

	my $objects = {};

	my @lines = split("\n", $body);
	my $obj;
	my $inObject = 0;
	foreach my $line (@lines)
	{
		if($line =~ /OK - Number of results: \d+/)
		{
			next;
		}
		if($line eq "")
		{
			$inObject = 0;
		}
		if($line =~ /^Name: (.+)/ and $inObject == 0)
		{
			debug("Found object: $1\n");
			$inObject = 1;
			$obj = {};
			my $index = encodeIndex($1);
			$objects->{$index} = $obj;
			$obj->{"index"} = $index;

		}
		elsif($line =~ /(.+?): (.+)/)
		{
			debug("Attribute $1: $2\n");
			$obj->{$1} = $2;

			if($1 eq "Usage")
			{
				parseUsage($obj,$2);
			}


			if($1 eq "objectName" and $2 =~ /Catalina:j2eeType=Servlet,name=(.+?),WebModule=(.+?),J2EEApplication.+/)
			{
				#Create a more human readable servlet name
				$obj->{"friendlyName"} = $2."/".$1;
				$obj->{"servletName"} = $1;
			}
		}
	}
	return $objects;
}

#
# Mempool usage is a complex attribute. Parse out the interesting parts
#

sub parseUsage
{
	my $obj = shift;
	my $usage = shift;
	$usage =~ /committed=(\d+)/;
	my $committed = $1;

	$usage =~ /init=(\d+)/;
	my $init = $1;

	$usage =~ /max=(\d+)/;
	my $max = $1;

	$usage =~ /used=(\d+)/;
	my $used = $1;

	$obj->{"Usage_Committed"} = $committed;
	$obj->{"Usage_Init"} = $init;
	$obj->{"Usage_Max"} = $max;
	$obj->{"Usage_Used"} = $used;

}



#
# Encode indeces as some indeces contain spaces.
#
sub encodeIndex
{
	my $name = shift;
	$name =~ s/ /%20/g;
	return $name;
}

#
# Decode indecex
#
sub decodeIndex
{
	my $name = shift;
	$name =~ s/%20/ /g;
	return $name;
}



#Attributes of interest
#======================
#gc (java.lang:type=GarbageCollector,name=*)
#
#CollectionTime
#CollectionCount
#Name

#tp (Catalina:type=ThreadPool,name=*)
#
#name
#currentThreadCount
#maxThreads
#currentThreadsBusy

#mp (java.lang:type=MemoryPool,name=*)
#
#Name
#Usage_Committed
#Usage_Init
#Usage_Max
#Usage_Used

#wm (Catalina:j2eeType=WebModule,name=*,J2EEApplication=none,J2EEServer=none)
#
# path
# processingTime
#

#sv (Catalina:j2eeType=Servlet,name=*,WebModule=*,J2EEApplication=none,J2EEServer=none)
#
# objectName
# friendlyName  Human readable name of servlet. Not a JMX attribute. This is something thins script creates.
# servletName  Short Human readable name of servlet. Not a JMX attribute. This is something thins script creates.
# processingTime Total execution time of the servlet's service method (millis)
# requestCount
# errorCount Error Count

